//
//  ContrastiveDivergence.cpp
//  
//
//  Created by Roni Mittelman on 7/24/13.
//
//

#include "ContrastiveDivergence.h"
#include <af/utils.h>
using namespace af;

array ContrastiveDivergence::sigma_function(array x){
    return (1.0/(1.0+exp(-x)));
}

array ContrastiveDivergence::computPH(array v){
    const unsigned MiniBatchLen=v.dims(0);
    array tm0=matmul(v,W).T();
    array tmp;
    if (alpha_parm.dims(0)==1){
        tmp=0.5/alpha_parm*tm0*tm0+matmul(v,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v*v,phi).T();
    }
    else{
        array O0=constant(1,NHidden,MiniBatchLen);
        tmp=0.5*O0/tile(alpha_parm,1,MiniBatchLen)*tm0*tm0+matmul(v,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v*v,phi).T();
    }
    array Eh_a=sigma_function(tmp);
    return Eh_a;
}

void ContrastiveDivergence::sampleCD_ssRBM(array v){
    dim4 dims = v.dims();
    const unsigned NVisible=dims[1];
    const unsigned MiniBatchLen=dims[0];
    v_a=v.copy();
    array tm0=matmul(v_a,W).T();
    array tmp;
    if (alpha_parm.dims(0)==1){
        tmp=0.5/alpha_parm*tm0*tm0+matmul(v_a,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_a*v_a,phi).T();
    }
    else{
        array O0=constant(1,NHidden,MiniBatchLen);
        tmp=0.5*O0/tile(alpha_parm,1,MiniBatchLen)*tm0*tm0+matmul(v_a,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_a*v_a,phi).T();
    }
    if (b.dims()[1]==1){
        tmp+=tile(flat(b),1,MiniBatchLen);
    }
    else{
        tmp+=b;
    }
    Eh_a=sigma_function(tmp);
    h_a=(Eh_a>randu(NHidden,MiniBatchLen)).as(f32);
    if (alpha_parm.dims(0)==1){
        Es_a=(matmul(v_a,W).T()/alpha_parm+tile(mu,1,MiniBatchLen))*h_a;
        s_a=Es_a+randn(NHidden,MiniBatchLen)/sqrt(alpha_parm);
    }
    else{
        Es_a=(matmul(v_a,W/tile(alpha_parm.T(),NVisible,1)).T()+tile(mu,1,MiniBatchLen))*h_a;
        s_a=Es_a+randn(NHidden,MiniBatchLen)/tile(sqrt(alpha_parm),1,MiniBatchLen);
    }
    array Ones0=constant(1,NVisible,MiniBatchLen);
    array CovMat;
    if (lambda_parm.dims(0)==1){
        CovMat=Ones0/(lambda_parm+matmul(phi,h_a));
    }
    else{
        CovMat=Ones0/(tile(lambda_parm,1,MiniBatchLen)+matmul(phi,h_a));
    }
    Ev_b=(CovMat*matmul(W,s_a*h_a)).T();
    v_b=Ev_b+randn(MiniBatchLen,NVisible)*sqrt(CovMat.T());

    int i0=1;
    while (1){
        tm0=matmul(v_b,W).T();
        if (alpha_parm.dims(0)==1){
            tmp=0.5/alpha_parm*tm0*tm0+matmul(v_b,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_b*v_b,phi).T();
        }
        else{
            array O0=constant(1,NHidden,MiniBatchLen);
            tmp=0.5*O0/tile(alpha_parm,1,MiniBatchLen)*tm0*tm0+matmul(v_b,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_b*v_b,phi).T();
        }
        if (b.dims()[1]==1){
            tmp+=tile(flat(b),1,MiniBatchLen);
        }
        else{
            tmp+=b;
        }
        Eh_b=sigma_function(tmp);

        h_b=(Eh_b>randu(NHidden,MiniBatchLen)).as(f32);
        if (alpha_parm.dims(0)==1){
            Es_b=(matmul(v_b,W).T()/alpha_parm+tile(mu,1,MiniBatchLen))*h_b;
            s_b=Es_b+randn(NHidden,MiniBatchLen)/sqrt(alpha_parm);
        }
        else{
            Es_b=(matmul(v_b,W/tile(alpha_parm.T(),NVisible,1)).T()+tile(mu,1,MiniBatchLen))*h_b;
            s_b=Es_b+randn(NHidden,MiniBatchLen)/tile(sqrt(alpha_parm),1,MiniBatchLen);
        }
        if (i0<n_CD){
            i0++;
            if (lambda_parm.dims(0)==1){
                CovMat=Ones0/(lambda_parm+matmul(phi,h_b));
            }
            else{
                CovMat=Ones0/(tile(lambda_parm,1,MiniBatchLen)+matmul(phi,h_b));
            }
            Ev_b=(CovMat*matmul(W,s_b*h_b)).T();
            v_b=Ev_b+randn(MiniBatchLen,NVisible)*sqrt(CovMat.T());
        }
        else{
            break;
        }
    }
}

void ContrastiveDivergence::sampleCDV(array v, array v0, array indices0){
    dim4 dims = v.dims();
    const unsigned NVisible=dims[1];
    const unsigned MiniBatchLen=dims[0];
    v_a=v.copy();
    array tm0=matmul(v_a,W).T();
    array tmp;
    if (alpha_parm.dims(0)==1){
        tmp=0.5/alpha_parm*tm0*tm0+matmul(v_a,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_a*v_a,phi).T();
    }
    else{
        array O0=constant(1,NHidden,MiniBatchLen);
        tmp=0.5*O0/tile(alpha_parm,1,MiniBatchLen)*tm0*tm0+matmul(v_a,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_a*v_a,phi).T();
    }
    if (b.dims()[1]==1){
        tmp+=tile(flat(b),1,MiniBatchLen);
    }
    else{
        tmp+=b;
    }
    Eh_a=sigma_function(tmp);
    h_a=(Eh_a>randu(NHidden,MiniBatchLen)).as(f32);
    if (alpha_parm.dims(0)==1){
        Es_a=(matmul(v_a,W).T()/alpha_parm+tile(mu,1,MiniBatchLen))*h_a;
        s_a=Es_a+randn(NHidden,MiniBatchLen)/sqrt(alpha_parm);
    }
    else{
        Es_a=(matmul(v_a,W/tile(alpha_parm.T(),NVisible,1)).T()+tile(mu,1,MiniBatchLen))*h_a;
        s_a=Es_a+randn(NHidden,MiniBatchLen)/tile(sqrt(alpha_parm),1,MiniBatchLen);
    }
    array Ones0=constant(1,NVisible,MiniBatchLen);
    array CovMat;
    if (lambda_parm.dims(0)==1){
        CovMat=Ones0/(lambda_parm+matmul(phi,h_a));
    }
    else{
        CovMat=Ones0/(tile(lambda_parm,1,MiniBatchLen)+matmul(phi,h_a));
    }
    Ev_b=(CovMat*matmul(W,s_a*h_a)).T();
    v_b=Ev_b+randn(MiniBatchLen,NVisible)*sqrt(CovMat.T());
    Ev_b(indices0)=v0(indices0);
    v_b(indices0)=v0(indices0);
    
    int i0=1;
    while (1){
        tm0=matmul(v_b,W).T();
        if (alpha_parm.dims(0)==1){
            tmp=0.5/alpha_parm*tm0*tm0+matmul(v_b,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_b*v_b,phi).T();
        }
        else{
            array O0=constant(1,NHidden,MiniBatchLen);
            tmp=0.5*O0/tile(alpha_parm,1,MiniBatchLen)*tm0*tm0+matmul(v_b,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v_b*v_b,phi).T();
        }
        if (b.dims()[1]==1){
            tmp+=tile(flat(b),1,MiniBatchLen);
        }
        else{
            tmp+=b;
        }
        Eh_b=sigma_function(tmp);
        
        h_b=(Eh_b>randu(NHidden,MiniBatchLen)).as(f32);
        if (alpha_parm.dims(0)==1){
            Es_b=(matmul(v_b,W).T()/alpha_parm+tile(mu,1,MiniBatchLen))*h_b;
            s_b=Es_b+randn(NHidden,MiniBatchLen)/sqrt(alpha_parm);
        }
        else{
            Es_b=(matmul(v_b,W/tile(alpha_parm.T(),NVisible,1)).T()+tile(mu,1,MiniBatchLen))*h_b;
            s_b=Es_b+randn(NHidden,MiniBatchLen)/tile(sqrt(alpha_parm),1,MiniBatchLen);
        }
        if (i0<n_CD){
            i0++;
            if (lambda_parm.dims(0)==1){
                CovMat=Ones0/(lambda_parm+matmul(phi,h_b));
            }
            else{
                CovMat=Ones0/(tile(lambda_parm,1,MiniBatchLen)+matmul(phi,h_b));
            }
            Ev_b=(CovMat*matmul(W,s_b*h_b)).T();
            v_b=Ev_b+randn(MiniBatchLen,NVisible)*sqrt(CovMat.T());
            Ev_b(indices0)=v0(indices0);
            v_b(indices0)=v0(indices0);
        }
        else{
            break;
        }
    }
}

void ContrastiveDivergence::MeanFieldV(array vt, array v0, array indices0){
    array tmp;
    array v=vt;
    dim4 dims = v.dims();
    const unsigned MiniBatchLen=dims[0];
    int Nmf=n_CD;
    //////
    if (flag_binary_real_ssRBM==3){
        array tm0=matmul(v,W).T();
        if (alpha_parm.dims(0)==1){
            tmp=0.5/alpha_parm*tm0*tm0+matmul(v,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v*v,phi).T();
        }
        else{
            array O0=constant(1,NHidden,MiniBatchLen);
            tmp=0.5*O0/tile(alpha_parm,1,MiniBatchLen)*tm0*tm0+matmul(v,W).T()*tile(mu,1,MiniBatchLen)-0.5*matmul(v*v,phi).T();
        }
        if (b.dims()[1]==1){
            tmp+=tile(flat(b),1,MiniBatchLen);
        }
        else{
            tmp+=b;
        }
        Eh_a=sigma_function(tmp);
    }
    //    rand_init(v);
    //////
    for (int i=0;i<Nmf;i++){
        if (b.dims()[1]==1){
            if (flag_binary_real_ssRBM!=3){
                tmp = matmul(W.T(), v.T())+tile(flat(b),1,MiniBatchLen);
            }
            else{
                if (alpha_parm.dims(0)==1){
                    Es_a=(matmul(v,W).T()/alpha_parm+tile(mu,1,MiniBatchLen))*Eh_a;
                }
                else{
                    Es_a=(matmul(v,W).T()/tile(alpha_parm,1,MiniBatchLen)+tile(mu,1,MiniBatchLen))*Eh_a;
                }
                tmp=matmul(v,W).T()*Es_a-0.5*matmul(v*v,phi).T()+Es_a*tile(alpha_parm*mu,1,MiniBatchLen)+tile(flat(b),1,MiniBatchLen)-0.5*tile(alpha_parm*mu*mu,1,MiniBatchLen);
            }
        }
        else{
            if (flag_binary_real_ssRBM!=3){
                tmp = matmul(W.T(), v.T())+b;
            }
            else{
                if (alpha_parm.dims(0)==1){
                    Es_a=(matmul(v,W).T()/alpha_parm+tile(mu,1,MiniBatchLen))*Eh_a;
                }
                else{
                    Es_a=(matmul(v,W).T()/tile(alpha_parm,1,MiniBatchLen)+tile(mu,1,MiniBatchLen))*Eh_a;
                }
                tmp=matmul(v,W).T()*Es_a-0.5*matmul(v*v,phi).T()+Es_a*tile(alpha_parm*mu,1,MiniBatchLen)+b-0.5*tile(alpha_parm*mu*mu,1,MiniBatchLen);
            }
        }

        Eh_a=sigma_function(tmp);
        h_a=Eh_a;
        //////////
        if (flag_binary_real_ssRBM!=3){
            array prob_v=(matmul(W,Eh_a)+tile(flat(c),1,MiniBatchLen)).T();
            Ev_b=prob_v;
            if (flag_binary_real_ssRBM==0){
                Ev_b=sigma_function(Ev_b);
            }
        }
        else{
            array Ones0=constant(1,v.dims(0),MiniBatchLen);
            array CovMat;
            if (lambda_parm.dims(0)==1){
                CovMat=Ones0/(lambda_parm+matmul(phi,Eh_a));
            }
            else{
                CovMat=Ones0/(tile(lambda_parm,1,MiniBatchLen)+matmul(phi,Eh_a));
            }
            Ev_b=(CovMat*matmul(W,Es_a*Eh_a)).T();
        }
        Ev_b(indices0)=v0(indices0);
        v_b=Ev_b;
        v=v_b;
    }
}

void ContrastiveDivergence::sampleCD_binary_real(array v){
    dim4 dims = v.dims();
    const unsigned NVisible=dims[1];
    const unsigned MiniBatchLen=dims[0];

    v_a=v.copy();
    array tmp;
    if (b.dims()[1]==1){
        tmp = matmul(W.T(), v_a.T())+tile(flat(b),1,MiniBatchLen);
    }
    else{
        tmp = matmul(W.T(), v_a.T())+b;
    }
    
    Eh_a=sigma_function(tmp);
    h_a=(Eh_a>randu(NHidden,MiniBatchLen)).as(f32);
    
    if (flag_binary_real_ssRBM==0){
        array prob_v=sigma_function(matmul(W,h_a)+tile(flat(c),1,MiniBatchLen)).T();
        Ev_b=prob_v;
        v_b=(randu(MiniBatchLen, NVisible)<prob_v).as(f32);
    }
    else{
        if (flag_binary_real_ssRBM==2){
            array prob_v=(matmul(W,h_a)+tile(flat(c),1,MiniBatchLen)).T();
            Ev_b=prob_v;
            v_b=prob_v;
        }
    }
    int i0=1;
    while (1){
        if (b.dims()[1]==1){
            tmp = matmul(W.T(), v_b.T())+tile(flat(b),1,MiniBatchLen);
        }
        else{
            tmp = matmul(W.T(), v_b.T())+b;
        }
        Eh_b=sigma_function(tmp);

        h_b=(Eh_b>randu(NHidden,MiniBatchLen)).as(f32);
        if (i0<n_CD){
            i0++;
            if (flag_binary_real_ssRBM==0){
                array prob_v=sigma_function(matmul(W,h_b)+tile(flat(c),1,MiniBatchLen)).T();
                Ev_b=prob_v;
                v_b=(randu(MiniBatchLen, NVisible)<prob_v).as(f32);
            }
            else{
                if (flag_binary_real_ssRBM==2){
                    array prob_v=(matmul(W,h_b)+tile(flat(c),1,MiniBatchLen)).T();
                    Ev_b=prob_v;
                    v_b=prob_v;
                }
            }
        }
        else{
            break;
        }
    }
}


void ContrastiveDivergence::sampleCD(array v){
    if ((flag_binary_real_ssRBM==0) || (flag_binary_real_ssRBM==2)){
        sampleCD_binary_real(v);
    }
    else{
        sampleCD_ssRBM(v);
    }
}


void ContrastiveDivergence::updateParms(array W0, array b0, array c0, array phi0, array mu0, array alpha_parm0, array lambda_parm0){
    W=array(W0);
    b=array(b0);
    c=array(c0);
    phi=array(phi0);
    mu=array(mu0);
    alpha_parm=array(alpha_parm0);
    lambda_parm=array(lambda_parm0);
}

void ContrastiveDivergence::updateParms(array W0, array b0, array c0){
    W=array(W0);
    b=array(b0);
    c=array(c0);
}

array ContrastiveDivergence::get_h_a(){
    return h_a;
}
array ContrastiveDivergence::get_h_b(){
    return h_b;
}
array ContrastiveDivergence::get_v_a(){
    return v_a;
}
array ContrastiveDivergence::get_v_b(){
    return v_b;
}

array ContrastiveDivergence::get_Eh_a(){
    return Eh_a;
}

array ContrastiveDivergence::get_Eh_b(){
    return Eh_b;
}

array ContrastiveDivergence::get_s_a(){
    return s_a;
}

array ContrastiveDivergence::get_s_b(){
    return s_b;
}

array ContrastiveDivergence::get_Ev_b(){
    return Ev_b;
}


int ContrastiveDivergence::get_flag_binary_real_ssRBM(){
    return flag_binary_real_ssRBM;
}


ContrastiveDivergence::ContrastiveDivergence(int nCD0, int NHidden0, int flag_binary_real_ssRBM0){
    n_CD=nCD0;
    NHidden=NHidden0;
    flag_binary_real_ssRBM=flag_binary_real_ssRBM0;
}

array ContrastiveDivergence::getW(){
    return W;
}

void  ContrastiveDivergence::set_flags(int flag_binary_real_ssRBM0){
    flag_binary_real_ssRBM=flag_binary_real_ssRBM0;
}
void ContrastiveDivergence::set_NHidden(int NHidden0){
    NHidden=NHidden0;
}

ContrastiveDivergence::ContrastiveDivergence(){
    
}

ContrastiveDivergence::~ContrastiveDivergence(){
}